@@ -1,5 +1,5 @@
 //
-// Copyright 2010-2011 Ettus Research LLC
+// Copyright 2010-2011,2014 Ettus Research LLC
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 #include <curses.h>
 #include <iostream>
 #include <complex>
+#include <cstdlib>
 
 namespace po = boost::program_options;
 
@@ -55,6 +56,7 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
         ("ref-lvl", po::value<float>(&ref_lvl)->default_value(0), "reference level for the display (dB)")
         ("dyn-rng", po::value<float>(&dyn_rng)->default_value(60), "dynamic range for the display (dB)")
         ("ref", po::value<std::string>(&ref)->default_value("internal"), "waveform type (internal, external, mimo)")
+        ("int-n", "tune USRP with integer-N tuning")
     ;
     po::variables_map vm;
     po::store(po::parse_command_line(argc, argv, desc), vm);
@@ -63,7 +65,7 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
     //print the help message
     if (vm.count("help") or not vm.count("rate")){
         std::cout << boost::format("UHD RX ASCII Art DFT %s") % desc << std::endl;
-        return ~0;
+        return EXIT_FAILURE;
     }
 
     //create a usrp device
@@ -82,7 +84,7 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
     //set the sample rate
     if (not vm.count("rate")){
         std::cerr << "Please specify the sample rate with --rate" << std::endl;
-        return ~0;
+        return EXIT_FAILURE;
     }
     std::cout << boost::format("Setting RX Rate: %f Msps...") % (rate/1e6) << std::endl;
     usrp->set_rx_rate(rate);
@@ -91,10 +93,12 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
     //set the center frequency
     if (not vm.count("freq")){
         std::cerr << "Please specify the center frequency with --freq" << std::endl;
-        return ~0;
+        return EXIT_FAILURE;
     }
     std::cout << boost::format("Setting RX Freq: %f MHz...") % (freq/1e6) << std::endl;
-    usrp->set_rx_freq(freq);
+    uhd::tune_request_t tune_request(freq);
+    if(vm.count("int-n")) tune_request.args = uhd::device_addr_t("mode_n=integer");
+    usrp->set_rx_freq(tune_request);
     std::cout << boost::format("Actual RX Freq: %f MHz...") % (usrp->get_rx_freq()/1e6) << std::endl << std::endl;
 
     //set the rf gain
@@ -147,7 +151,7 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
     //-- Initialize
     //------------------------------------------------------------------
     initscr(); //curses init
-    usrp->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
+    rx_stream->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
     boost::system_time next_refresh = boost::get_system_time();
 
     //------------------------------------------------------------------
@@ -188,11 +192,11 @@ int UHD_SAFE_MAIN(int argc, char *argv[]){
     //------------------------------------------------------------------
     //-- Cleanup
     //------------------------------------------------------------------
-    usrp->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS);
+    rx_stream->issue_stream_cmd(uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS);
     endwin(); //curses done
 
     //finished
     std::cout << std::endl << "Done!" << std::endl << std::endl;
 
-    return 0;
+    return EXIT_SUCCESS;
 }
