@@ -21,7 +21,7 @@
 #include <uhd/config.hpp>
 #include <boost/utility.hpp>
 #include <boost/shared_ptr.hpp>
-#include "wb_iface.hpp"
+#include <uhd/types/wb_iface.hpp>
 #include <complex>
 #include <string>
 
@@ -29,13 +29,15 @@ class rx_frontend_core_200 : boost::noncopyable{
 public:
     typedef boost::shared_ptr<rx_frontend_core_200> sptr;
 
-    static sptr make(wb_iface::sptr iface, const size_t base);
+    static sptr make(uhd::wb_iface::sptr iface, const size_t base);
 
     virtual void set_mux(const bool swap) = 0;
 
-    virtual void set_offset(const std::complex<double> &off) = 0;
+    virtual void set_dc_offset_auto(const bool enb) = 0;
 
-    virtual void set_correction(const std::complex<double> &cor) = 0;
+    virtual std::complex<double> set_dc_offset(const std::complex<double> &off) = 0;
+
+    virtual void set_iq_balance(const std::complex<double> &cor) = 0;
 
 };
 
