@@ -26,6 +26,10 @@
 #include <boost/thread/thread.hpp>
 #include <boost/tokenizer.hpp>
 #include <boost/format.hpp>
+#include <boost/regex.hpp>
+
+#include "boost/tuple/tuple.hpp"
+#include "boost/foreach.hpp"
 
 using namespace uhd;
 using namespace boost::gregorian;
@@ -34,56 +38,114 @@ using namespace boost::algorithm;
 using namespace boost::this_thread;
 
 /*!
- * A GPS control for Jackson Labs devices (and other NMEA compatible GPS's)
+ * A control for GPSDO devices
  */
 
 class gps_ctrl_impl : public gps_ctrl{
+private:
+  std::map<std::string, boost::tuple<std::string, boost::system_time, bool> > sensors;
+
+  std::string get_cached_sensor(const std::string sensor, const int freshness, const bool once, const bool touch=true) {
+    boost::system_time time = boost::get_system_time();
+    try {
+      // this is nasty ...
+      //std::cout << boost::format("Requested %s - seen? ") % sensor << sensors[sensor].get<2>() << " once? " << once << std::endl;
+      if(time - sensors[sensor].get<1>() < milliseconds(freshness) && (!once or !sensors[sensor].get<2>())) {
+        sensors[sensor] = boost::make_tuple(sensors[sensor].get<0>(), sensors[sensor].get<1>(), touch);
+        return sensors[sensor].get<0>();
+      } else {
+          return update_cached_sensors(sensor);
+      }
+    } catch(std::exception &e) {
+      UHD_MSG(warning) << "get_cached_sensor: " << e.what();
+    }
+    return std::string();
+  }
+
+  std::string update_cached_sensors(const std::string sensor) {
+    if(not gps_detected() || (gps_type != GPS_TYPE_INTERNAL_GPSDO)) {
+        UHD_MSG(error) << "get_stat(): unsupported GPS or no GPS detected";
+        return std::string();
+    }
+
+    const std::list<std::string> list = boost::assign::list_of("GPGGA")("GPRMC")("SERVO");
+    static const boost::regex status_regex("\\d\\d-\\d\\d-\\d\\d");
+    std::map<std::string,std::string> msgs;
+
+    // Get all GPSDO messages available
+    // Creating a map here because we only want the latest of each message type
+    for (std::string msg = _recv(); msg.length() > 6; msg = _recv())
+    {
+        // Look for SERVO message
+        if (boost::regex_search(msg, status_regex, boost::regex_constants::match_continuous))
+            msgs["SERVO"] = msg;
+        else
+            msgs[msg.substr(1,5)] = msg;
+    }
+
+    boost::system_time time = boost::get_system_time();
+
+    // Update sensors with newly read data
+    BOOST_FOREACH(std::string key, list) {
+        if (msgs[key].length())
+            sensors[key] = boost::make_tuple(msgs[key], time, !sensor.compare(key));
+    }
+
+    // Return requested sensor if it was updated
+    if (msgs[sensor].length())
+        return msgs[sensor];
+
+    return std::string();
+  }
+
 public:
   gps_ctrl_impl(uart_iface::sptr uart){
     _uart = uart;
 
+
     std::string reply;
     bool i_heard_some_nmea = false, i_heard_something_weird = false;
     gps_type = GPS_TYPE_NONE;
-    
-    //first we look for a Jackson Labs Firefly (since that's what we provide...)
+
+    //first we look for an internal GPSDO
     _flush(); //get whatever junk is in the rx buffer right now, and throw it away
-    _send("HAAAY GUYYYYS\n"); //to elicit a response from the Firefly
+    _send("HAAAY GUYYYYS\n"); //to elicit a response from the GPSDO
 
     //wait for _send(...) to return
-    sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+    sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
 
     //then we loop until we either timeout, or until we get a response that indicates we're a JL device
     const boost::system_time comm_timeout = boost::get_system_time() + milliseconds(GPS_COMM_TIMEOUT_MS);
     while(boost::get_system_time() < comm_timeout) {
       reply = _recv();
       if(reply.find("Command Error") != std::string::npos) {
-        gps_type = GPS_TYPE_JACKSON_LABS;
+        gps_type = GPS_TYPE_INTERNAL_GPSDO;
         break;
-      } 
+      }
       else if(reply.substr(0, 3) == "$GP") i_heard_some_nmea = true; //but keep looking for that "Command Error" response
       else if(reply.length() != 0) i_heard_something_weird = true; //probably wrong baud rate
       sleep(milliseconds(GPS_TIMEOUT_DELAY_MS));
     }
 
-    if((i_heard_some_nmea) && (gps_type != GPS_TYPE_JACKSON_LABS)) gps_type = GPS_TYPE_GENERIC_NMEA;
+    if((i_heard_some_nmea) && (gps_type != GPS_TYPE_INTERNAL_GPSDO)) gps_type = GPS_TYPE_GENERIC_NMEA;
 
     if((gps_type == GPS_TYPE_NONE) && i_heard_something_weird) {
       UHD_MSG(error) << "GPS invalid reply \"" << reply << "\", assuming none available" << std::endl;
     }
 
     switch(gps_type) {
-    case GPS_TYPE_JACKSON_LABS:
-      UHD_MSG(status) << "Found a Jackson Labs GPS" << std::endl;
-      init_firefly();
+    case GPS_TYPE_INTERNAL_GPSDO:
+      UHD_MSG(status) << "Found an internal GPSDO" << std::endl;
+      init_gpsdo();
       break;
 
     case GPS_TYPE_GENERIC_NMEA:
-      if(gps_type == GPS_TYPE_GENERIC_NMEA) UHD_MSG(status) << "Found a generic NMEA GPS device" << std::endl;
+      UHD_MSG(status) << "Found a generic NMEA GPS device" << std::endl;
       break;
 
     case GPS_TYPE_NONE:
     default:
+      UHD_MSG(status) << "No GPSDO found" << std::endl;
       break;
 
     }
@@ -99,7 +161,8 @@ public:
         ("gps_gpgga")
         ("gps_gprmc")
         ("gps_time")
-        ("gps_locked");
+        ("gps_locked")
+        ("gps_servo");
     return ret;
   }
 
@@ -108,7 +171,7 @@ public:
     or key == "gps_gprmc") {
         return sensor_value_t(
                  boost::to_upper_copy(key),
-                 get_nmea(boost::to_upper_copy(key.substr(4,8))),
+                 get_cached_sensor(boost::to_upper_copy(key.substr(4,8)), GPS_NMEA_NORMAL_FRESHNESS, false, false),
                  "");
     }
     else if(key == "gps_time") {
@@ -117,49 +180,52 @@ public:
     else if(key == "gps_locked") {
         return sensor_value_t("GPS lock status", locked(), "locked", "unlocked");
     }
+    else if(key == "gps_servo") {
+        return sensor_value_t("GPS servo status", get_servo(), "");
+    }
     else {
         throw uhd::value_error("gps ctrl get_sensor unknown key: " + key);
     }
   }
 
 private:
-  void init_firefly(void) {
+  void init_gpsdo(void) {
     //issue some setup stuff so it spits out the appropriate data
     //none of these should issue replies so we don't bother looking for them
     //we have to sleep between commands because the JL device, despite not acking, takes considerable time to process each command.
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
     _send("SYST:COMM:SER:ECHO OFF\n");
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
     _send("SYST:COMM:SER:PRO OFF\n");
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
     _send("GPS:GPGGA 1\n");
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
     _send("GPS:GGAST 0\n");
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
     _send("GPS:GPRMC 1\n");
-     sleep(milliseconds(FIREFLY_STUPID_DELAY_MS));
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
+    _send("SERV:TRAC 0\n");
+     sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
   }
- 
+
   //retrieve a raw NMEA sentence
   std::string get_nmea(std::string msgtype) {
-    msgtype.insert(0, "$");
     std::string reply;
-    if(not gps_detected()) {
-        UHD_MSG(error) << "get_nmea(): unsupported GPS or no GPS detected";
-        return std::string();
-    }
-
-    _flush(); //flush all input before waiting for a message
 
     const boost::system_time comm_timeout = boost::get_system_time() + milliseconds(GPS_COMM_TIMEOUT_MS);
     while(boost::get_system_time() < comm_timeout) {
-        reply = _recv();
-        if(reply.substr(0, 6) == msgtype)
-          return reply;
+        if(!msgtype.compare("GPRMC")) {
+          reply = get_cached_sensor(msgtype, GPS_NMEA_FRESHNESS, true);
+        }
+        else {
+          reply = get_cached_sensor(msgtype, GPS_NMEA_LOW_FRESHNESS, false);
+        }
+        if(reply.size()) {
+          if(reply.substr(1, 5) == msgtype) return reply;
+        }
         boost::this_thread::sleep(milliseconds(GPS_TIMEOUT_DELAY_MS));
     }
     throw uhd::value_error(str(boost::format("get_nmea(): no %s message found") % msgtype));
-    return std::string();
   }
 
   //helper function to retrieve a field from an NMEA sentence
@@ -176,9 +242,10 @@ private:
   }
 
   ptime get_time(void) {
+    _flush();
     int error_cnt = 0;
     ptime gps_time;
-    while(error_cnt < 3) {
+    while(error_cnt < 2) {
         try {
             std::string reply = get_nmea("GPRMC");
 
@@ -188,29 +255,30 @@ private:
             if(datestr.size() == 0 or timestr.size() == 0) {
                 throw uhd::value_error(str(boost::format("Invalid response \"%s\"") % reply));
             }
-            
+
             //just trust me on this one
-            gps_time = ptime( date( 
+            gps_time = ptime( date(
                              greg_year(boost::lexical_cast<int>(datestr.substr(4, 2)) + 2000),
-                             greg_month(boost::lexical_cast<int>(datestr.substr(2, 2))), 
-                             greg_day(boost::lexical_cast<int>(datestr.substr(0, 2))) 
+                             greg_month(boost::lexical_cast<int>(datestr.substr(2, 2))),
+                             greg_day(boost::lexical_cast<int>(datestr.substr(0, 2)))
                            ),
                           hours(  boost::lexical_cast<int>(timestr.substr(0, 2)))
                         + minutes(boost::lexical_cast<int>(timestr.substr(2, 2)))
                         + seconds(boost::lexical_cast<int>(timestr.substr(4, 2)))
                      );
             return gps_time;
-            
+
         } catch(std::exception &e) {
             UHD_MSG(warning) << "get_time: " << e.what();
+            _flush();
             error_cnt++;
         }
     }
     throw uhd::value_error("Timeout after no valid message found");
-    
+
     return gps_time; //keep gcc from complaining
   }
-  
+
   time_t get_epoch_time(void) {
       return (get_time() - from_time_t(0)).total_seconds();
   }
@@ -223,7 +291,7 @@ private:
     int error_cnt = 0;
     while(error_cnt < 3) {
         try {
-            std::string reply = get_nmea("GPGGA");
+            std::string reply = get_cached_sensor("GPGGA", GPS_LOCK_FRESHNESS, false, false);
             if(reply.size() <= 1) return false;
 
             return (get_token(reply, 6) != "0");
@@ -236,6 +304,28 @@ private:
     return false;
   }
 
+  std::string get_servo(void) {
+
+    //enable servo reporting
+    _send("SERV:TRAC 1\n");
+    sleep(milliseconds(GPSDO_STUPID_DELAY_MS));
+
+    std::string reply;
+
+    const boost::system_time comm_timeout = boost::get_system_time() + milliseconds(GPS_COMM_TIMEOUT_MS);
+    while(boost::get_system_time() < comm_timeout) {
+        reply = get_cached_sensor("SERVO", GPS_NMEA_LOW_FRESHNESS, false);
+        if(reply.size())
+        {
+            //disable it before leaving function
+            _send("SERV:TRAC 0\n");
+            return reply;
+        }
+        boost::this_thread::sleep(milliseconds(GPS_TIMEOUT_DELAY_MS));
+    }
+    throw uhd::value_error("get_stat(): no servo message found");
+  }
+
   uart_iface::sptr _uart;
 
   void _flush(void){
@@ -244,8 +334,8 @@ private:
     }
   }
 
-  std::string _recv(void){
-      return _uart->read_uart(GPS_TIMEOUT_DELAY_MS/1000.);
+  std::string _recv(double timeout = GPS_TIMEOUT_DELAY_MS/1000.){
+      return _uart->read_uart(timeout);
   }
 
   void _send(const std::string &buf){
@@ -253,14 +343,19 @@ private:
   }
 
   enum {
-    GPS_TYPE_JACKSON_LABS,
+    GPS_TYPE_INTERNAL_GPSDO,
     GPS_TYPE_GENERIC_NMEA,
     GPS_TYPE_NONE
   } gps_type;
 
-  static const int GPS_COMM_TIMEOUT_MS = 1500;
+  static const int GPS_COMM_TIMEOUT_MS = 1300;
+  static const int GPS_NMEA_FRESHNESS = 10;
+  static const int GPS_NMEA_LOW_FRESHNESS = 2500;
+  static const int GPS_NMEA_NORMAL_FRESHNESS = 1000;
+  static const int GPS_SERVO_FRESHNESS = 2500;
+  static const int GPS_LOCK_FRESHNESS = 2500;
   static const int GPS_TIMEOUT_DELAY_MS = 200;
-  static const int FIREFLY_STUPID_DELAY_MS = 200;
+  static const int GPSDO_STUPID_DELAY_MS = 200;
 };
 
 /***********************************************************************
