@@ -1,5 +1,5 @@
 //
-// Copyright 2010-2012 Ettus Research LLC
+// Copyright 2010-2013 Ettus Research LLC
 //
 // This program is free software: you can redistribute it and/or modify
 // it under the terms of the GNU General Public License as published by
@@ -19,11 +19,12 @@
 #include <uhd/usrp/multi_usrp.hpp>
 #include <uhd/utils/msg.hpp>
 #include <uhd/exception.hpp>
-#include <uhd/utils/msg.hpp>
+#include <uhd/utils/log.hpp>
 #include <uhd/utils/gain_group.hpp>
 #include <uhd/usrp/dboard_id.hpp>
 #include <uhd/usrp/mboard_eeprom.hpp>
 #include <uhd/usrp/dboard_eeprom.hpp>
+#include <uhd/convert.hpp>
 #include <boost/assign/list_of.hpp>
 #include <boost/thread.hpp>
 #include <boost/foreach.hpp>
@@ -35,6 +36,20 @@ using namespace uhd::usrp;
 
 const std::string multi_usrp::ALL_GAINS = "";
 
+UHD_INLINE std::string string_vector_to_string(std::vector<std::string> values, std::string delimiter = std::string(" "))
+{
+    std::string out = "";
+    for (std::vector<std::string>::iterator iter = values.begin(); iter != values.end(); iter++)
+    {
+        out += (iter != values.begin() ? delimiter : "") + *iter;
+    }
+    return out;
+}
+
+#define THROW_GAIN_NAME_ERROR(name,chan,dir) throw uhd::exception::runtime_error( \
+            (boost::format("%s: gain \"%s\" not found for channel %d.\nAvailable gains: %s\n") % \
+            __FUNCTION__ % name % chan % string_vector_to_string(get_##dir##_gain_names(chan))).str());
+
 /***********************************************************************
  * Helper methods
  **********************************************************************/
@@ -89,6 +104,8 @@ static meta_range_t make_overall_tune_range(
     return range;
 }
 
+
+
 /***********************************************************************
  * Gain helper functions
  **********************************************************************/
@@ -139,6 +156,14 @@ static tune_result_t tune_xx_subdev_and_dsp(
         if (bw > rate) lo_offset = std::min((bw - rate)/2, rate/2);
     }
 
+    //------------------------------------------------------------------
+    //-- poke the tune request args into the dboard
+    //------------------------------------------------------------------
+    if (rf_fe_subtree->exists("tune_args"))
+    {
+        rf_fe_subtree->access<device_addr_t>("tune_args").set(tune_request.args);
+    }
+
     //------------------------------------------------------------------
     //-- set the RF frequency depending upon the policy
     //------------------------------------------------------------------
@@ -501,6 +526,46 @@ public:
         return _tree->access<std::vector<std::string> >(mb_root(mboard) / "clock_source" / "options").get();
     }
 
+    void set_clock_source_out(const bool enb, const size_t mboard)
+    {
+        if (mboard != ALL_MBOARDS)
+        {
+            if (_tree->exists(mb_root(mboard) / "clock_source" / "output"))
+            {
+                _tree->access<bool>(mb_root(mboard) / "clock_source" / "output").set(enb);
+            }
+            else
+            {
+                throw uhd::runtime_error("multi_usrp::set_clock_source_out - not supported on this device");
+            }
+            return;
+        }
+        for (size_t m = 0; m < get_num_mboards(); m++)
+        {
+            this->set_clock_source_out(enb, m);
+        }
+    }
+
+    void set_time_source_out(const bool enb, const size_t mboard)
+    {
+        if (mboard != ALL_MBOARDS)
+        {
+            if (_tree->exists(mb_root(mboard) / "time_source" / "output"))
+            {
+                _tree->access<bool>(mb_root(mboard) / "time_source" / "output").set(enb);
+            }
+            else
+            {
+                throw uhd::runtime_error("multi_usrp::set_time_source_out - not supported on this device");
+            }
+            return;
+        }
+        for (size_t m = 0; m < get_num_mboards(); m++)
+        {
+            this->set_time_source_out(enb, m);
+        }
+    }
+
     size_t get_num_mboards(void){
         return _tree->list("/mboards").size();
     }
@@ -527,6 +592,11 @@ public:
     /*******************************************************************
      * RX methods
      ******************************************************************/
+    rx_streamer::sptr get_rx_stream(const stream_args_t &args) {
+        _check_link_rate(args, false);
+        return this->get_device()->get_rx_stream(args);
+    }
+
     void set_rx_subdev_spec(const subdev_spec_t &spec, size_t mboard){
         if (mboard != ALL_MBOARDS){
             _tree->access<subdev_spec_t>(mb_root(mboard) / "rx_subdev_spec").set(spec);
@@ -537,8 +607,25 @@ public:
         }
     }
 
-    subdev_spec_t get_rx_subdev_spec(size_t mboard){
-        return _tree->access<subdev_spec_t>(mb_root(mboard) / "rx_subdev_spec").get();
+    subdev_spec_t get_rx_subdev_spec(size_t mboard)
+    {
+        subdev_spec_t spec = _tree->access<subdev_spec_t>(mb_root(mboard) / "rx_subdev_spec").get();
+        if (spec.empty())
+        {
+            try
+            {
+                const std::string db_name = _tree->list(mb_root(mboard) / "dboards").at(0);
+                const std::string fe_name = _tree->list(mb_root(mboard) / "dboards" / db_name / "rx_frontends").at(0);
+                spec.push_back(subdev_spec_pair_t(db_name, fe_name));
+                _tree->access<subdev_spec_t>(mb_root(mboard) / "rx_subdev_spec").set(spec);
+            }
+            catch(const std::exception &e)
+            {
+                throw uhd::index_error(str(boost::format("multi_usrp::get_rx_subdev_spec(%u) failed to make default spec - %s") % mboard % e.what()));
+            }
+            UHD_MSG(status) << "Selecting default RX front end spec: " << spec.to_pp_string() << std::endl;
+        }
+        return spec;
     }
 
     size_t get_rx_num_channels(void){
@@ -595,15 +682,27 @@ public:
     }
 
     void set_rx_gain(double gain, const std::string &name, size_t chan){
-        return rx_gain_group(chan)->set_value(gain, name);
+        try {
+            return rx_gain_group(chan)->set_value(gain, name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,rx);
+        }
     }
 
     double get_rx_gain(const std::string &name, size_t chan){
-        return rx_gain_group(chan)->get_value(name);
+        try {
+            return rx_gain_group(chan)->get_value(name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,rx);
+        }
     }
 
     gain_range_t get_rx_gain_range(const std::string &name, size_t chan){
-        return rx_gain_group(chan)->get_range(name);
+        try {
+            return rx_gain_group(chan)->get_range(name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,rx);
+        }
     }
 
     std::vector<std::string> get_rx_gain_names(size_t chan){
@@ -679,6 +778,11 @@ public:
     /*******************************************************************
      * TX methods
      ******************************************************************/
+    tx_streamer::sptr get_tx_stream(const stream_args_t &args) {
+        _check_link_rate(args, true);
+        return this->get_device()->get_tx_stream(args);
+    }
+
     void set_tx_subdev_spec(const subdev_spec_t &spec, size_t mboard){
         if (mboard != ALL_MBOARDS){
             _tree->access<subdev_spec_t>(mb_root(mboard) / "tx_subdev_spec").set(spec);
@@ -689,8 +793,25 @@ public:
         }
     }
 
-    subdev_spec_t get_tx_subdev_spec(size_t mboard){
-        return _tree->access<subdev_spec_t>(mb_root(mboard) / "tx_subdev_spec").get();
+    subdev_spec_t get_tx_subdev_spec(size_t mboard)
+    {
+        subdev_spec_t spec = _tree->access<subdev_spec_t>(mb_root(mboard) / "tx_subdev_spec").get();
+        if (spec.empty())
+        {
+            try
+            {
+                const std::string db_name = _tree->list(mb_root(mboard) / "dboards").at(0);
+                const std::string fe_name = _tree->list(mb_root(mboard) / "dboards" / db_name / "tx_frontends").at(0);
+                spec.push_back(subdev_spec_pair_t(db_name, fe_name));
+                _tree->access<subdev_spec_t>(mb_root(mboard) / "tx_subdev_spec").set(spec);
+            }
+            catch(const std::exception &e)
+            {
+                throw uhd::index_error(str(boost::format("multi_usrp::get_tx_subdev_spec(%u) failed to make default spec - %s") % mboard % e.what()));
+            }
+            UHD_MSG(status) << "Selecting default TX front end spec: " << spec.to_pp_string() << std::endl;
+        }
+        return spec;
     }
 
     size_t get_tx_num_channels(void){
@@ -747,15 +868,27 @@ public:
     }
 
     void set_tx_gain(double gain, const std::string &name, size_t chan){
-        return tx_gain_group(chan)->set_value(gain, name);
+        try {
+            return tx_gain_group(chan)->set_value(gain, name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,tx);
+        }
     }
 
     double get_tx_gain(const std::string &name, size_t chan){
-        return tx_gain_group(chan)->get_value(name);
+        try {
+            return tx_gain_group(chan)->get_value(name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,tx);
+        }
     }
 
     gain_range_t get_tx_gain_range(const std::string &name, size_t chan){
-        return tx_gain_group(chan)->get_range(name);
+        try {
+            return tx_gain_group(chan)->get_range(name);
+        } catch (uhd::key_error &e) {
+            THROW_GAIN_NAME_ERROR(name,chan,tx);
+        }
     }
 
     std::vector<std::string> get_tx_gain_names(size_t chan){
@@ -818,6 +951,74 @@ public:
         }
     }
 
+    /*******************************************************************
+     * GPIO methods
+     ******************************************************************/
+    std::vector<std::string> get_gpio_banks(const size_t mboard)
+    {
+        std::vector<std::string> banks;
+        if (_tree->exists(mb_root(mboard) / "gpio"))
+        {
+            BOOST_FOREACH(const std::string &name, _tree->list(mb_root(mboard) / "gpio"))
+            {
+                banks.push_back(name);
+            }
+        }
+        BOOST_FOREACH(const std::string &name, _tree->list(mb_root(mboard) / "dboards"))
+        {
+            banks.push_back("RX"+name);
+            banks.push_back("TX"+name);
+        }
+        return banks;
+    }
+
+    void set_gpio_attr(const std::string &bank, const std::string &attr, const boost::uint32_t value, const boost::uint32_t mask, const size_t mboard)
+    {
+        if (_tree->exists(mb_root(mboard) / "gpio" / bank))
+        {
+            const boost::uint32_t current = _tree->access<boost::uint32_t>(mb_root(mboard) / "gpio" / bank / attr).get();
+            const boost::uint32_t new_value = (current & ~mask) | (value & mask);
+            _tree->access<boost::uint32_t>(mb_root(mboard) / "gpio" / bank / attr).set(new_value);
+            return;
+        }
+        if (bank.size() > 2 and bank[1] == 'X')
+        {
+            const std::string name = bank.substr(2);
+            const dboard_iface::unit_t unit = (bank[0] == 'R')? dboard_iface::UNIT_RX : dboard_iface::UNIT_TX;
+            dboard_iface::sptr iface = _tree->access<dboard_iface::sptr>(mb_root(mboard) / "dboards" / name / "iface").get();
+            if (attr == "CTRL") iface->set_pin_ctrl(unit, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "DDR") iface->set_gpio_ddr(unit, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "OUT") iface->set_gpio_out(unit, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "ATR_0X") iface->set_atr_reg(unit, dboard_iface::ATR_REG_IDLE, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "ATR_RX") iface->set_atr_reg(unit, dboard_iface::ATR_REG_RX_ONLY, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "ATR_TX") iface->set_atr_reg(unit, dboard_iface::ATR_REG_TX_ONLY, boost::uint16_t(value), boost::uint16_t(mask));
+            if (attr == "ATR_XX") iface->set_atr_reg(unit, dboard_iface::ATR_REG_FULL_DUPLEX, boost::uint16_t(value), boost::uint16_t(mask));
+        }
+    }
+
+    boost::uint32_t get_gpio_attr(const std::string &bank, const std::string &attr, const size_t mboard)
+    {
+        if (_tree->exists(mb_root(mboard) / "gpio" / bank))
+        {
+            return _tree->access<boost::uint64_t>(mb_root(mboard) / "gpio" / bank / attr).get();
+        }
+        if (bank.size() > 2 and bank[1] == 'X')
+        {
+            const std::string name = bank.substr(2);
+            const dboard_iface::unit_t unit = (bank[0] == 'R')? dboard_iface::UNIT_RX : dboard_iface::UNIT_TX;
+            dboard_iface::sptr iface = _tree->access<dboard_iface::sptr>(mb_root(mboard) / "dboards" / name / "iface").get();
+            if (attr == "CTRL") return iface->get_pin_ctrl(unit);
+            if (attr == "DDR") return iface->get_gpio_ddr(unit);
+            if (attr == "OUT") return iface->get_gpio_out(unit);
+            if (attr == "ATR_0X") return iface->get_atr_reg(unit, dboard_iface::ATR_REG_IDLE);
+            if (attr == "ATR_RX") return iface->get_atr_reg(unit, dboard_iface::ATR_REG_RX_ONLY);
+            if (attr == "ATR_TX") return iface->get_atr_reg(unit, dboard_iface::ATR_REG_TX_ONLY);
+            if (attr == "ATR_XX") return iface->get_atr_reg(unit, dboard_iface::ATR_REG_FULL_DUPLEX);
+            if (attr == "READBACK") return iface->read_gpio(unit);
+        }
+        return 0;
+    }
+
 private:
     device::sptr _dev;
     property_tree::sptr _tree;
@@ -835,6 +1036,10 @@ private:
             if (mcp.chan < sss) break;
             mcp.chan -= sss;
         }
+        if (mcp.mboard >= get_num_mboards())
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp: RX channel %u out of range for configured RX frontends") % chan));
+        }
         return mcp;
     }
 
@@ -846,48 +1051,119 @@ private:
             if (mcp.chan < sss) break;
             mcp.chan -= sss;
         }
+        if (mcp.mboard >= get_num_mboards())
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp: TX channel %u out of range for configured TX frontends") % chan));
+        }
         return mcp;
     }
 
-    fs_path mb_root(const size_t mboard){
-        const std::string name = _tree->list("/mboards").at(mboard);
-        return "/mboards/" + name;
+    fs_path mb_root(const size_t mboard)
+    {
+        try
+        {
+            const std::string name = _tree->list("/mboards").at(mboard);
+            return "/mboards/" + name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::mb_root(%u) - %s") % mboard % e.what()));
+        }
     }
 
-    fs_path rx_dsp_root(const size_t chan){
+    fs_path rx_dsp_root(const size_t chan)
+    {
         mboard_chan_pair mcp = rx_chan_to_mcp(chan);
-        const std::string name = _tree->list(mb_root(mcp.mboard) / "rx_dsps").at(mcp.chan);
-        return mb_root(mcp.mboard) / "rx_dsps" / name;
+        if (_tree->exists(mb_root(mcp.mboard) / "rx_chan_dsp_mapping")) {
+            std::vector<size_t> map = _tree->access<std::vector<size_t> >(mb_root(mcp.mboard) / "rx_chan_dsp_mapping").get();
+            UHD_ASSERT_THROW(map.size() > mcp.chan);
+            mcp.chan = map[mcp.chan];
+        }
+
+        try
+        {
+            const std::string name = _tree->list(mb_root(mcp.mboard) / "rx_dsps").at(mcp.chan);
+            return mb_root(mcp.mboard) / "rx_dsps" / name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::rx_dsp_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
-    fs_path tx_dsp_root(const size_t chan){
+    fs_path tx_dsp_root(const size_t chan)
+    {
         mboard_chan_pair mcp = tx_chan_to_mcp(chan);
-        const std::string name = _tree->list(mb_root(mcp.mboard) / "tx_dsps").at(mcp.chan);
-        return mb_root(mcp.mboard) / "tx_dsps" / name;
+        if (_tree->exists(mb_root(mcp.mboard) / "tx_chan_dsp_mapping")) {
+            std::vector<size_t> map = _tree->access<std::vector<size_t> >(mb_root(mcp.mboard) / "tx_chan_dsp_mapping").get();
+            UHD_ASSERT_THROW(map.size() > mcp.chan);
+            mcp.chan = map[mcp.chan];
+        }
+        try
+        {
+            const std::string name = _tree->list(mb_root(mcp.mboard) / "tx_dsps").at(mcp.chan);
+            return mb_root(mcp.mboard) / "tx_dsps" / name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::tx_dsp_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
-    fs_path rx_fe_root(const size_t chan){
+    fs_path rx_fe_root(const size_t chan)
+    {
         mboard_chan_pair mcp = rx_chan_to_mcp(chan);
-        const subdev_spec_pair_t spec = get_rx_subdev_spec(mcp.mboard).at(mcp.chan);
-        return mb_root(mcp.mboard) / "rx_frontends" / spec.db_name;
+        try
+        {
+            const subdev_spec_pair_t spec = get_rx_subdev_spec(mcp.mboard).at(mcp.chan);
+            return mb_root(mcp.mboard) / "rx_frontends" / spec.db_name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::rx_fe_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
-    fs_path tx_fe_root(const size_t chan){
+    fs_path tx_fe_root(const size_t chan)
+    {
         mboard_chan_pair mcp = tx_chan_to_mcp(chan);
-        const subdev_spec_pair_t spec = get_tx_subdev_spec(mcp.mboard).at(mcp.chan);
-        return mb_root(mcp.mboard) / "tx_frontends" / spec.db_name;
+        try
+        {
+            const subdev_spec_pair_t spec = get_tx_subdev_spec(mcp.mboard).at(mcp.chan);
+            return mb_root(mcp.mboard) / "tx_frontends" / spec.db_name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::tx_fe_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
-    fs_path rx_rf_fe_root(const size_t chan){
+    fs_path rx_rf_fe_root(const size_t chan)
+    {
         mboard_chan_pair mcp = rx_chan_to_mcp(chan);
-        const subdev_spec_pair_t spec = get_rx_subdev_spec(mcp.mboard).at(mcp.chan);
-        return mb_root(mcp.mboard) / "dboards" / spec.db_name / "rx_frontends" / spec.sd_name;
+        try
+        {
+            const subdev_spec_pair_t spec = get_rx_subdev_spec(mcp.mboard).at(mcp.chan);
+            return mb_root(mcp.mboard) / "dboards" / spec.db_name / "rx_frontends" / spec.sd_name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::rx_rf_fe_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
-    fs_path tx_rf_fe_root(const size_t chan){
+    fs_path tx_rf_fe_root(const size_t chan)
+    {
         mboard_chan_pair mcp = tx_chan_to_mcp(chan);
-        const subdev_spec_pair_t spec = get_tx_subdev_spec(mcp.mboard).at(mcp.chan);
-        return mb_root(mcp.mboard) / "dboards" / spec.db_name / "tx_frontends" / spec.sd_name;
+        try
+        {
+            const subdev_spec_pair_t spec = get_tx_subdev_spec(mcp.mboard).at(mcp.chan);
+            return mb_root(mcp.mboard) / "dboards" / spec.db_name / "tx_frontends" / spec.sd_name;
+        }
+        catch(const std::exception &e)
+        {
+            throw uhd::index_error(str(boost::format("multi_usrp::tx_rf_fe_root(%u) - mcp(%u) - %s") % chan % mcp.chan % e.what()));
+        }
     }
 
     gain_group::sptr rx_gain_group(size_t chan){
@@ -915,11 +1191,40 @@ private:
         }
         return gg;
     }
+
+    //! \param is_tx True for tx
+    // Assumption is that all mboards use the same link
+    bool _check_link_rate(const stream_args_t &args, bool is_tx) {
+        bool link_rate_is_ok = true;
+        size_t bytes_per_sample = convert::get_bytes_per_item(args.otw_format.empty() ? "sc16" : args.otw_format);
+        double max_link_rate = 0;
+        double sum_rate = 0;
+        BOOST_FOREACH(const size_t chan, args.channels) {
+            mboard_chan_pair mcp = is_tx ? tx_chan_to_mcp(chan) : rx_chan_to_mcp(chan);
+            if (_tree->exists(mb_root(mcp.mboard) / "link_max_rate")) {
+                max_link_rate = std::max(
+                    max_link_rate,
+                   _tree->access<double>(mb_root(mcp.mboard) / "link_max_rate").get()
+                );
+            }
+            sum_rate += is_tx ? get_tx_rate(chan) : get_rx_rate(chan);
+        }
+        if (max_link_rate > 0 and (max_link_rate / bytes_per_sample) < sum_rate) {
+            UHD_MSG(warning) << boost::format(
+                "The total sum of rates (%f MSps on %u channels) exceeds the maximum capacity of the connection.\n"
+                "This can cause %s."
+            ) % (sum_rate/1e6) % args.channels.size() % (is_tx ? "underruns (U)" : "overflows (O)")  << std::endl;
+            link_rate_is_ok = false;
+        }
+
+        return link_rate_is_ok;
+    }
 };
 
 /***********************************************************************
  * The Make Function
  **********************************************************************/
 multi_usrp::sptr multi_usrp::make(const device_addr_t &dev_addr){
+    UHD_LOG << "multi_usrp::make with args " << dev_addr.to_pp_string() << std::endl;
     return sptr(new multi_usrp_impl(dev_addr));
 }
